/*****************************************************************************\
 *
 * Midpoint3.cpp
 *
 * Midpoint algorithm for terrain height map
 * 
 *  Description: Terrain
 *
 *      Authors: PCJohn (peciva _AT fit.vutbr.cz)
 *               
 * Contributors:
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/

#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoEnvironment.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoNormalBinding.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include <stdlib.h>
#include <time.h>
//

// #include "cmd_parser.h"

SoSeparator* createTerrain(int argc, char **argv);  // Funkce vytvejc scnu

// Velikost vkov mapy zvten o jednu
// (krajina 64x64 se konstruuje ze 65x65 bod)
#define DEFAULT_TERRAIN_DIMENSION  65

// Velikost generovan krajiny v prostoru
#define DEFAULT_TERRAIN_SIZE 64.f 

// Maximln vka a hloubka krajiny
#define DEFAULT_TERRAIN_HEIGHT_Q  20.f//15.f

// Filtrovac koeficient pro finln vyhlazen krajiny.
// Vy slo odpovd vy "hrbolatosti",
// ni - vce se to vyfiltruje.
#define DEFAULT_TERRAIN_FILTER_Q  0.8f // 0.5f

// HQ (high quality) uruje, jakm zpsobem bude rendrovna vsledn krajina.
// HQ=0 znamen, e kad tverec bude rendrovn jako dva trojhelnky,
// HQ=1 je podobn pedchoz volb, ale zlom mezi tmito trojhelnky je
// vytven podle toho, ve kterm smru je v dan dladici vt zmna vky,
// HQ=2 pid do stedu tverce dal vertex a vyrendruje tverec tymi
// trojhelnky. Krajina bude mt dvojnsobn mnostv trojhelnk,
// ale stane se hlad.
#define DEFAULT_TERRAIN_HQ  2 //0



/*!
  Tda pedstavujc tverec v krajin.
*/
class TerrainPatch {
private:
  void allocHmap(int dimension);
  void freeHmap();
  void filter(float q);
  SbVec3f calculateNormal(const int x, const int y, const float sx, const float sy);

protected:
  float *hmap;
  int dimension;

  inline int getHmapIndex(const int x, const int y);
  int getMaterialIndex(const float height, const float minHeight, const float maxHeight);

public:
  TerrainPatch();
  ~TerrainPatch();

  void generateHeightMap();
  void generateHeightMap(int dimension, float heightQ, float filterQ);

  SoSeparator* getModel();
  SoSeparator* getModel(float size, int hq);

  static SoSeparator* generate();
  static SoSeparator* generate(int dimension, float size, float heightQ, float filterQ, int hq);

  void reset();
};



/*!
  Funkce vrac index do vkov mapy hmap pro souadnice [x,y].
*/
inline int TerrainPatch::getHmapIndex(const int x, const int y)
{
  return y*dimension + x;
}



// Funkce vracejc nhodnou float hodnotu v rozsahu -max a +max
float randf(float max)
{
  return (float(rand()) / float(RAND_MAX)) * max*2 - max;
}



/*!
  Funkce vytvo graf scny reprezentujc tverec krajiny.
  Parametry krajiny jsou pouity standardn.
*/
SoSeparator* TerrainPatch::generate()
{
  TerrainPatch terrain;
  terrain.generateHeightMap();
  return terrain.getModel();
}



/*!
  Funkce vytvo graf scny reprezentujc tverec krajiny.
*/
SoSeparator* TerrainPatch::generate(int dimension, float size, float heightQ, float filterQ, int hq)
{
  TerrainPatch terrain;
  terrain.generateHeightMap(dimension, heightQ, filterQ);
  return terrain.getModel(size, hq);
}



/*!
  Tato funkce vytvo metodou pesouvn prostednho bodu
  vkovou mapu, kter je uloena do promnn hmap.
  Parametr dimension udv rozmr vkov mapy, kter mus bt tvercov.
  Rozmr mus bt mocninou 2 plus jedna. Tedy pro krajinu sloenou
  ze 64x64 ploek musme vytvoit vkovou mapu o rozmrech 65x65.
  Parametr heightQ udv vkovou lenitost krajiny - "jak bude vysok".
  filterQ se udv pro zvren profiltrovn hrbolatosti povrchu,
  kter znan eliminuje viditelnost hran mezi trojhelnky.
*/
void TerrainPatch::generateHeightMap(int dimension, float heightQ, float filterQ)
{
  // Zkontroluj platnost parametr danch uivatelem.
  // dimension mus bt mocnina 2 plus 1
  assert(dimension >= 1 && "Terrain dimension must be greater than zero.");
  assert(int(pow(2, int(log(dimension-1)/log(2.f)+0.5f))+0.5f) == dimension-1 &&
         "Terrain dimension is not power of 2 plus 1");
  // filterQ mus bt v rozsahu od 0.f do 1.f
  assert(filterQ >= 0.f || filterQ <= 1.f &&
         "Wrong filter coeficient");

  // Alokujeme vkovou mapu
  allocHmap(dimension);

  // Pia poten vku do roh mapy
  hmap[getHmapIndex(          0,           0)] = 0.f;
  hmap[getHmapIndex(dimension-1,           0)] = 0.f;
  hmap[getHmapIndex(          0, dimension-1)] = 0.f;
  hmap[getHmapIndex(dimension-1, dimension-1)] = 0.f;

  // V nsledujc smyce se generuj postupn vy a vy detaily
  // mapy. Zan se s nejvtm krokem (promnn step nastavena na
  // velikost pole), a kon se s krokem 2. Po t ji znme vky
  // vech bod.
  int step = dimension-1;
  float maxR = heightQ;
  int x,y;
  while (step >= 2) {

    // 1. fze: tvercov mka vodorovn
    // Ze ty bod tverce vygeneruj prostedn bod a k jeho hodnot
    // piti nhodn slo v rozsahu -maxR do +maxR
    for (y=0; y<dimension-1; y+=step) {
      for (x=0; x<dimension-1; x+=step) {
        hmap[getHmapIndex(x+step/2, y+step/2)] =
            (hmap[getHmapIndex(     x, y)] +
             hmap[getHmapIndex(x+step, y)] +
             hmap[getHmapIndex(     x, y+step)] +
             hmap[getHmapIndex(x+step, y+step)]) / 4.f + randf(maxR);
      }
    }

    // Korekce nhodnho rozptylu vky
    maxR *= sqrtf(0.5f);

    // 2. fze: tvercov mka pootoen o 45 stup
    // Ze ty bod opt vygeneruj prostedn bod a jeho hodnotu
    // modifikuj. Po dokonen tto fze je hustota platnch
    // platnch dat v na vkov map dvojnsobn.
    for (y=0; y<dimension-1; y+=step) {
      for (x=0; x<dimension-1; x+=step) {
        hmap[getHmapIndex(x+step/2, y)] =
            ((y-step/2 >= 0 ?
             hmap[getHmapIndex(x+step/2, y-step/2)] : 0) +
             hmap[getHmapIndex(x       , y       )] +
             hmap[getHmapIndex(x+step  , y       )] +
             hmap[getHmapIndex(x+step/2, y+step/2)]) /
             (y-step/2 >= 0 ? 4 : 3) + randf(maxR);
        hmap[getHmapIndex(x, y+step/2)] =
            (hmap[getHmapIndex(x       , y       )] +
             (x-step/2 >= 0 ?
             hmap[getHmapIndex(x-step/2, y+step/2)] : 0) +
             hmap[getHmapIndex(x+step/2, y+step/2)] +
             hmap[getHmapIndex(x       , y+step  )]) /
             (x-step/2 >= 0 ? 4 : 3) + randf(maxR);
        hmap[getHmapIndex(x+step, y+step/2)] =
            (hmap[getHmapIndex(x+step       , y       )] +
             hmap[getHmapIndex(x+step/2     , y+step/2)] +
             (x+step+step/2 < dimension ?
             hmap[getHmapIndex(x+step+step/2, y+step/2)] : 0) +
             hmap[getHmapIndex(x+step       , y+step  )]) /
             (x+step+step/2 < dimension ? 4 : 3) + randf(maxR);
        hmap[getHmapIndex(x+step/2, y+step)] =
            (hmap[getHmapIndex(x+step/2, y+step/2     )] +
             hmap[getHmapIndex(x       , y+step       )] +
             hmap[getHmapIndex(x+step  , y+step       )] +
             (y+step+step/2 < dimension ?
             hmap[getHmapIndex(x+step/2, y+step+step/2)] : 0)) /
             (y+step+step/2 < dimension ? 4 : 3) + randf(maxR);
      }
    }

    // Korekce nhodnho rozptylu vky
    maxR *= sqrtf(0.5f);

    // Zmeni krok na polovinu
    step /= 2;
  }

  // Zvren vyfiltrovn povrchu.
  // (Toto filtrovn podstatn vylep vzhled povrchu,
  // pokud je povrch svtlovn njakm svtlem.)
  filter(filterQ);
}



/*!
  Tato funkce je ekvivalentn funkci generateMap(int, float, float),
  piem pouije defaultn parametry.
*/
void TerrainPatch::generateHeightMap()
{
  generateHeightMap(DEFAULT_TERRAIN_DIMENSION, DEFAULT_TERRAIN_HEIGHT_Q,
                    DEFAULT_TERRAIN_FILTER_Q);
}



/*!
  Funkce vytvo zobraziteln graf scny modelu krajiny
  zaloen na vkov map. Pokud dosud nebyla vkov mapa
  vytvoena, vytvo se volnm funkce getHeightMap().
  Parametr size udv velikost krajiny (v prostorovch souadnicch)
  a parametr hq udv metodu vytvoen modelu krajiny.
  0 zna standardn tesselaci, tj. dva trojhelnky na tverec krajiny,
  1 je podobn pedchoz volb, ale zlom mezi tmito trojhelnky je
  vytven podle toho, ve kterm smru je v dan dladici vt zmna vky,
  2 dladice je rendrovna tymi trojhelnky.
*/
SoSeparator* TerrainPatch::getModel(float size, int hq)
{
  // generate height map if not already generated
  if (!hmap)
    generateHeightMap();

  // Vpoet minHeight a maxHeight
  float minHeight, maxHeight;
  minHeight = maxHeight = hmap[0];
  int i;
  for (i=1; i<dimension*dimension; i++) {
    if (hmap[i] < minHeight)
      minHeight = hmap[i];
    if (hmap[i] > maxHeight)
      maxHeight = hmap[i];
  }

  // Vytvo koenov nod
  SoSeparator *root = new SoSeparator;

  // Vytvo model vodn hladiny
  // Nejprve materil, pak souadnice, a na zvr stripSet
  SoMaterial *waterMat = new SoMaterial;
  waterMat->ambientColor.setValue(0.f, 0.f, 1.f);
  waterMat->diffuseColor.setValue(0.f, 0.f, 0.f);
  root->addChild(waterMat);

  SoCoordinate3 *waterCoords = new SoCoordinate3;
  float waterLevel = (maxHeight - minHeight) / 5.f + minHeight;
  waterCoords->point.set1Value(0, SbVec3f(0.f,  waterLevel, 0.f));
  waterCoords->point.set1Value(1, SbVec3f(size, waterLevel, 0.f));
  waterCoords->point.set1Value(2, SbVec3f(0.f,  waterLevel, size));
  waterCoords->point.set1Value(3, SbVec3f(size, waterLevel, size));
  root->addChild(waterCoords);

  SoTriangleStripSet *waterStripSet = new SoTriangleStripSet;
  waterStripSet->numVertices.set1Value(0, 4); // nult prvek nastav na hodnotu 4,
      // to znamen, e budeme renderovat 2 trojhelnky (poet bod snme o
      // dva a zskme tak poet trojhelnk
  root->addChild(waterStripSet);

  // Materil (barva) krajiny,
  // rzn indexy odpovdaj rznm barvm ternu.
  SoMaterial *material = new SoMaterial;
  material->ambientColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
  material->diffuseColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
  material->ambientColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
  material->diffuseColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
  material->ambientColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
  material->diffuseColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
  material->ambientColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
  material->diffuseColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
  root->addChild(material);

  // Pro normly nastavujeme binding, tj. zda-li specifikujeme
  // normly pro kad bod (PER_VERTEX), nebo jen pro trojhelnky
  // (PER_FACE), i cel plocha m jen jednu normlu (OVERALL).
  // Ppona _INDEXED znamen, e se normly nebudou brt sekvenn
  // z nodu SoNormal, nbr se se budou indexovat ze
  // SoIndexedTriangleStripSet::normalIndex podobn jako se
  // souadnice indexuj z SoIndexedTriangleStripSet::coordIndex.
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);
  root->addChild(nbind);

  // Binding pro materily,
  // vce informac u pedchozho komente k SoNormalBinding.
  SoMaterialBinding *mbind = new SoMaterialBinding;
  mbind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
  root->addChild(mbind);

  // Vytvome nody a nastavme poet souadnic a norml, kter
  // budeme generovat. (template SbSqr pot druhou mocninu)
  SoCoordinate3 *coords = new SoCoordinate3;
  SoNormal *normals = new SoNormal;
  int num = (hq != 2) ? SbSqr(dimension) : SbSqr(dimension)+SbSqr(dimension-1);
  coords->point.setNum(num);
  normals->vector.setNum(num);
  SbVec3f *c = coords->point.startEditing();
  SbVec3f *ctemp = c;
  SbVec3f *n = normals->vector.startEditing();


  // Nyn generujeme trojhelnky do pamti. Kad tverec z na
  // vkov mapy renderujeme jako tyi trojhelnky. Teoreticky
  // by staily dva, ale model by pi osvtlovn vykazoval chyby.
  int xi,yi;
  float sx = size / (dimension-1);
  float sy = sx;
  SbVec3f midp;

  if (hq != 2) {
    for (yi=0; yi<dimension; yi++) {
      for (xi=0; xi<dimension; xi++) {

        // souadnice a normla pro lev spodn roh tverce v krajin
        *(c++) = SbVec3f(float(xi)*sx, hmap[getHmapIndex(xi,yi)], float(yi)*sy);
        *(n++) = calculateNormal(xi, yi, sx, sy);

      }
    }
  } else {
    for (yi=0; yi<dimension-1; yi++) {
      for (xi=0; xi<dimension-1; xi++) {

        // Vpoet prostednho bodu tverce
        midp.setValue((float(xi)+0.5f) * sx,
                      (hmap[getHmapIndex(xi,yi)] + hmap[getHmapIndex(xi+1,yi)] +
                      hmap[getHmapIndex(xi,yi+1)]+ hmap[getHmapIndex(xi+1,yi+1)]) / 4.f,
                      (float(yi)+0.5f) * sy);

        // Spotme si normly pro tyi rohov body i pro
        // stedov bod
        SbVec3f n1 = calculateNormal(xi, yi, sx, sy);
        SbVec3f n2 = calculateNormal(xi, yi+1, sx, sy);
        SbVec3f n3 = calculateNormal(xi+1, yi+1, sx, sy);
        SbVec3f n4 = calculateNormal(xi+1, yi, sx, sy);
        SbVec3f nmid = n1+n2+n3+n4;
        nmid.normalize();

        // souadnice a normla pro lev spodn roh tverce v krajin
        *(c++) = SbVec3f(float(xi)*sx, hmap[getHmapIndex(xi,yi)], float(yi)*sy);
        *(n++) = n1;

        // souadnice a normla pro prostedn bod ve tverci krajiny
        *(c++) = midp;
        *(n++) = nmid;
      }
      // posledn bod v dku
      *(c++) = SbVec3f(float(xi)*sx, hmap[getHmapIndex(xi,yi)], float(yi)*sy);
      *(n++) = calculateNormal(xi, yi, sx, sy);
    }

    // Posledn ada vertex
    for (xi=0; xi<dimension; xi++) {
      *(c++) = SbVec3f(float(xi)*sx, hmap[getHmapIndex(xi,yi)], float(yi)*sy);
      *(n++) = calculateNormal(xi, yi, sx, sy);
    }
  }


  // Nakonfiguruj triangleStripSet
  SoIndexedTriangleStripSet *stripSet = new SoIndexedTriangleStripSet;
  switch (hq) {
  case 0: num = dimension * (dimension-1) * 2 + dimension-1; break;
  case 1: num = dimension * (dimension-1) * 5; break;
  case 2: num = (dimension-1) * (dimension-1) * 10; break;
  }
  stripSet->coordIndex.setNum(num);
  stripSet->materialIndex.setNum(num);

  // Prvnmu ze strip vdy pia 5 souadnic (vytvo 3 trojhelnky)
  // a druhmu 3 (vytvo jeden trojhelnk).
  int32_t *ci = stripSet->coordIndex.startEditing();
  int32_t *mi = stripSet->materialIndex.startEditing();
  switch (hq) {
  case 0: {
    c = ctemp;
    int i1 = 0;
    int i2 = dimension;
    for (yi=0; yi<dimension-1; yi++) {
      for (xi=0; xi<dimension; xi++) {
        *(ci++) = i2;   // lev horn bod tverce
        *(ci++) = i1;   // lev spodn bod tverce
        // Materilov indexy pro vertexy, kter jsme vygenerovali
        // na pedchozch dcch.
        *(mi++) = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // lev horn bod tverce
        *(mi++) = getMaterialIndex(c[i1][1], minHeight, maxHeight);   // lev spodn bod tverce
        i1++;
        i2++;
      }
      *(ci++) = -1;     // konec triangle stripu
      *(mi++) = -1;
    }
    break;
  }
  case 1: {
    c = ctemp;
    int i1 = 0;
    int i2 = dimension;
    for (yi=0; yi<dimension-1; yi++) {
      for (xi=0; xi<dimension-1; xi++) {
        // Potej zmnu v jednotlivch smrech dladice,
        // abychom zjistili, ve kterm smru bude dladice rozdlena.
        bool orderChange = fabsf(c[i2+1][1]-c[i1][1]) > fabsf(c[i2][1]-c[i1+1][1]);
        if (orderChange) {
          int tmp = i2;
          i2 = i1;
          i1 = tmp;
        }

        *(ci++) = i2;   // lev horn bod tverce
        *(ci++) = i1;   // lev spodn bod tverce
        // Materilov indexy pro vertexy, kter jsme vygenerovali
        // na pedchozch dcch.
        *(mi++) = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // lev horn bod tverce
        *(mi++) = getMaterialIndex(c[i1][1], minHeight, maxHeight);   // lev spodn bod tverce
        i1++;
        i2++;
        *(ci++) = i2;   // prav horn bod tverce
        *(ci++) = i1;   // prav spodn bod tverce
        *(mi++) = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // prav horn bod tverce
        *(mi++) = getMaterialIndex(c[i1][1], minHeight, maxHeight);   // prav spodn bod tverce
        *(ci++) = -1;
        *(mi++) = -1;
      }
      i1++;
      i2++;
    }
    break;
  }
  case 2: {
    c = ctemp;
    int tmp1,tmp2,tmp3;
    int i1 = 0;
    int i2 = 2*dimension-1;
    for (yi=0; yi<dimension-2; yi++) {
      for (xi=0; xi<dimension-1; xi++) {
        // Generovn souadnicovch index
        // (vdy dva triangleStripSety, prvn sloen ze t,
        // druh z jednoho trojhelnku)
        *(ci++) = i2;   // lev horn bod tverce
        *(ci++) = i1;   // lev spodn bod tverce
        *(ci++) = i1+1; // prostedn bod
        *(ci++) = i1+2; // prav spodn bod
        *(ci++) = i2+2; // prav horn bod
        *(ci++) = -1;
        *(ci++) = i2;   // lev horn bod
        *(ci++) = i1+1; // prostedn bod
        *(ci++) = i2+2; // prav horn bod
        *(ci++) = -1;

        // Materilov indexy pro vertexy, kter jsme vygenerovali
        // na pedchozch dcch.
        *(mi++) = tmp1 = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // lev horn bod tverce
        *(mi++) =        getMaterialIndex(c[i1][1], minHeight, maxHeight);   // lev spodn bod tverce
        *(mi++) = tmp2 = getMaterialIndex(c[i1+1][1], minHeight, maxHeight); // prostedn bod
        *(mi++) =        getMaterialIndex(c[i1+2][1], minHeight, maxHeight); // prav spodn bod
        *(mi++) = tmp3 = getMaterialIndex(c[i2+2][1], minHeight, maxHeight); // prav horn bod
        *(mi++) = -1;
        *(mi++) = tmp1; // lev horn bod
        *(mi++) = tmp2; // prostedn bod
        *(mi++) = tmp3; // prav horn bod
        *(mi++) = -1;

        i1 += 2;
        i2 += 2;
      }
      i1++;
      i2++;
    }

    // Posledn dek
    for (xi=0; xi<dimension-1; xi++) {
      *(ci++) = i2;
      *(ci++) = i1;
      *(ci++) = i1+1;
      *(ci++) = i1+2;
      *(ci++) = i2+1; // zde zmna oproti pedchoz smyce
      *(ci++) = -1;
      *(ci++) = i2;
      *(ci++) = i1+1;
      *(ci++) = i2+1; // zde zmna
      *(ci++) = -1;

      *(mi++) = tmp1 = getMaterialIndex(c[i2][1], minHeight, maxHeight);
      *(mi++) =        getMaterialIndex(c[i1][1], minHeight, maxHeight);
      *(mi++) = tmp2 = getMaterialIndex(c[i1+1][1], minHeight, maxHeight);
      *(mi++) =        getMaterialIndex(c[i1+2][1], minHeight, maxHeight);
      *(mi++) = tmp3 = getMaterialIndex(c[i2+1][1], minHeight, maxHeight); // zde zmna
      *(mi++) = -1;
      *(mi++) = tmp1;
      *(mi++) = tmp2;
      *(mi++) = tmp3;
      *(mi++) = -1;

      i1 += 2;
      i2 += 1;
    }
    break;
  }}
  stripSet->coordIndex.finishEditing();
  stripSet->materialIndex.finishEditing();
  coords->point.finishEditing();
  normals->vector.finishEditing();
  root->addChild(normals);
  root->addChild(coords);
  root->addChild(stripSet);

  return root;
}



/*!
  Funkce vytvo zobraziteln graf scny modelu krajiny
  zaloen na vkov map. Pokud dosud nebyla vkov mapa
  vytvoena, vytvo se volnm funkce getHeightMap().
  Parametry vyrven modelu jsou pouity standardn.
 */
SoSeparator* TerrainPatch::getModel()
{
  return getModel(DEFAULT_TERRAIN_SIZE, DEFAULT_TERRAIN_HQ);
}



/*!
  Funkce vyfiltruje vkovou mapu.
  Toto filtrovn znan eliminuje viditelnost hran mezi trojhelnky.
 */
void TerrainPatch::filter(float q)
{
  int x, y;

  // dky zleva doprava
  for (y = 0; y < dimension; y++)
    for (x = 1; x < dimension; x++)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x-1,y)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;

  // dky zprava doleva
  for (y = 0; y < dimension; y++)
    for (x = dimension-1; x >= 1; x--)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x-1,y)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;

  // sloupce zespoda nahoru
  for (y = 1; y < dimension; y++)
    for (x = 0; x < dimension; x++)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x,y-1)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;

  // sloupce shora dol
  for (y = dimension-1; y >= 1; y--)
    for (x = 0; x < dimension; x++)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x,y-1)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;
}



/*!
  Funkce alokuje novou vkovou mapu, kter je vdy
  tvercov. Jej rozmr je zadn v parametru dimension,
  kter mus bt mocninou 2 plus 1.
 */
void TerrainPatch::allocHmap(int dimension)
{
  if (this->hmap == NULL || this->dimension != dimension) {
    delete[] this->hmap;
    this->dimension = dimension;
    this->hmap = new float[dimension*dimension];
  }
}



/*!
  Funkce uvoln pam, kter byla alokovna pro vkovou mapu.
 */
void TerrainPatch::freeHmap()
{
  delete[] this->hmap;
  this->hmap = NULL;
}



/*!
  Funkce vypot normlu v danm bod vkov mapy.
 */
SbVec3f TerrainPatch::calculateNormal(const int x, const int y, const float sx, const float sy)
{
  float dx,dy;

  // hel v ose X
  if (x>=1 && x<=dimension-2)
    // Vpoet ze dvou sousednch bod
    dx = (hmap[getHmapIndex(x-1,y)] - hmap[getHmapIndex(x+1,y)]) / 2.f;
  else
    // Vpoet na okraji mapy
    if (x<1)  dx = hmap[getHmapIndex(x,y)] - hmap[getHmapIndex(x+1,y)];
    else  dx = hmap[getHmapIndex(x-1,y)] - hmap[getHmapIndex(x,y)];

  // hel v ose Y
  if (y>=1 && y<=dimension-2)
    dy = (hmap[getHmapIndex(x,y-1)] - hmap[getHmapIndex(x,y+1)]) / 2.f;
  else
    if (y<1)  dy = hmap[getHmapIndex(x,y)] - hmap[getHmapIndex(x,y+1)];
    else  dy = hmap[getHmapIndex(x,y-1)] - hmap[getHmapIndex(x,y)];


#if 0 // Neoptimalizovan kd, kter vyuv goniometrick funkce,
      // kter jsou vpoetn velmi nron.

  // Pevod na hel v radinech
  float angx = float(atanf(dx/sx));
  float angy = float(atanf(dy/sy));

  // Vpoet vslednho vektoru
  return SbVec3f(sinf(angx), cosf(angx)*cosf(angy), sinf(angy));
#else

  SbVec3f r(dx/sx, 1.f, dy/sy);
  r.normalize();
  return r;
#endif
}



/*!
  Funkce vrt index materilu pro dan bod vkov mapy.
 */
int TerrainPatch::getMaterialIndex(const float height, const float minHeight, const float maxHeight)
{
  float part = (maxHeight-minHeight) / 5.f;
  if (height <= minHeight+3.f*part)  return 1; // zelen materil
  if (height <= minHeight+4.f*part)  return 2; // utrovit materil
  return 3; // snh
}



/*!
  Konstruktor.
*/
TerrainPatch::TerrainPatch()
{
  hmap = NULL;
}



/*!
  Destruktor.
*/
TerrainPatch::~TerrainPatch()
{
  reset();
}



/*!
  Inicializuje objekt do podobnho stavu jako pi vytvoen
  objektu, tzn. minimln vyuit zdroj.
  Funkce se pouv pedevm pro uvolnn pamti
  vkov mapy, nen-li ji dle potebn.
*/
void TerrainPatch::reset()
{
  freeHmap();
}





// Funkce vytvo graf scny pro nai aplikaci.
SoSeparator* createTerrain(int argc, char **argv)
{
  int i = 1;

  // Inicializace genertoru nhodnch sel
  srand((unsigned int)time(NULL));

  // Vytvo koen scny
  SoSeparator *root = new SoSeparator;
  root->ref();

  // Move the object
  SoTranslation *t = new SoTranslation;
  t->translation.setValue(-45.f, 60.f, -50.0f);
  root->addChild(t);


  // Rotate
  SoRotationXYZ *rotX = new SoRotationXYZ;
  rotX->axis = SoRotationXYZ::X;
  rotX->angle.setValue(90.0f);
  root->addChild(rotX);

  // Nastav intenzitu globlnho svtla ve scn
  SoEnvironment *env = new SoEnvironment;
  env->ambientIntensity.setValue(0.1f);
  root->addChild(env);

  // Svtlo
  SoDirectionalLight *sunLight = new SoDirectionalLight;
  sunLight->direction.setValue(0.35f, -1.0f, 0.25f);
  sunLight->intensity.setValue(0.5f);
  root->addChild(sunLight);

  // Model krajiny
  int dimension = DEFAULT_TERRAIN_DIMENSION;
  float size = DEFAULT_TERRAIN_SIZE;
  float heightQ = DEFAULT_TERRAIN_HEIGHT_Q;
  float filterQ = DEFAULT_TERRAIN_FILTER_Q;
  int hq = DEFAULT_TERRAIN_HQ;

  // Omez hq na hodnoty 0,1 a 2
  if (hq < 0 || hq > 2)  hq = 2;
/*
  // Generovn krajiny
  printf("Generuji krajinu midpoint algoritmem:\n"
         "hmap dimension:  %i\n"
         "size:            %.1f\n"
         "heightQ:         %.1f\n"
         "filterQ:         %.2f\n"
         "hq:              %i\n",
         dimension, size, heightQ, filterQ, hq); 
*/
  root->addChild(TerrainPatch::generate(dimension, size, heightQ, filterQ, hq));

  root->unrefNoDelete();
  return root;
}
