/*****************************************************************************\
 *
 * scene.cpp
 *
 * Algorithm for shadows ganeration in a scene
 * 
 *  Description: Functions used to create the scene and user callbacks
 *               
 *      Authors: Tom Burian (xburia10 _AT stud.fit.vutbr.cz)
 *               
 *    Thanks to: PCJohn (peciva _AT fit.vutbr.cz)
 * 
 * Contributors:
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/

#include "scene.h"
#include "shadows_win.h"
#include "shadow_volume.h"
#include "scene_lights.h"
#include "cmd_parser.h"

extern SoSeparator* createTerrain(int argc, char **argv);

//
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

// use a z-fail algorithm; if false z-pass is used instead
int zfail = false;

// demo scenes
int demo = 1;


// Surface coordinates
static float surfaceVertices[8][3] =
{
  -SURFACE_SIZE2, -SURFACE_SIZE2, 0,  // bottom
   SURFACE_SIZE2, -SURFACE_SIZE2, 0,

   SURFACE_SIZE2,  SURFACE_SIZE2, 0,  // back
  -SURFACE_SIZE2,  SURFACE_SIZE2, 0,

  -SURFACE_SIZE2,  SURFACE_SIZE2, 2*SURFACE_SIZE2,  // right
   SURFACE_SIZE2,  SURFACE_SIZE2, 2*SURFACE_SIZE2,

  -SURFACE_SIZE2, -SURFACE_SIZE2, 2*SURFACE_SIZE2,  // left
   SURFACE_SIZE2, -SURFACE_SIZE2, 2*SURFACE_SIZE2,
};


// Surface indices
static int32_t surfaceIndices[8][4] =
{
  0, 1, 2, -1, // bottom
  3, 0, 2, -1,

  2, 5, 4, -1, // back
  3, 2, 4, -1,

  3, 4, 6, -1, // left
  0, 3, 6, -1,

  2, 1, 7, -1, // right
  5, 2, 7, -1,

};


//----------------------------------------------------------------------------
// Computes the triangle normal vector from its verteces.
// Counterclockwise orientation
//
SbVec3f faceNormal(const SbVec3f p1, const SbVec3f p2, const SbVec3f p3)
{
  SbVec3f vector1, vector2;

  vector1 = p2 - p1;
  vector2 = p3 - p2;

  SbVec3f normal = vector1.cross(vector2);
  normal.normalize();

  return normal;
}


//----------------------------------------------------------------------------
// Creates surface at depth 'depth' with 'color'
//
SoSeparator* createSurface(float depth, SbColor color)
{
  // sceneRoot node
  SoSeparator *sceneRoot = new SoSeparator;

  // Material
  SoMaterial *material = new SoMaterial;
  material->ambientColor.setValue(color);
  material->diffuseColor.setValue(color);
  sceneRoot->addChild(material);

  // Translation to set depth
  SoTranslation *trans = new SoTranslation;
  trans->translation.setValue(0.f, 0.f, depth);
  sceneRoot->addChild(trans);

  // Coordinates
  SoCoordinate3 *coords = new SoCoordinate3;
  coords->point.setValues(0, 8, surfaceVertices);
  sceneRoot->addChild(coords);

  for (int i=0; i<7; i++) {
    SoIndexedTriangleStripSet *strip = new SoIndexedTriangleStripSet;
    strip->coordIndex.setNum(1);
    strip->coordIndex.setValues(0, 8, surfaceIndices[i]);
    sceneRoot->addChild(strip);
  }

  return sceneRoot;
}


//----------------------------------------------------------------------------
// Keyboard callback
//
void myKeyPressCallback (void * userdata, SoEventCallback * keyboardCallback)
{
  ShadowVolume * sv = (ShadowVolume *) userdata;  

  const SoEvent * event = keyboardCallback->getEvent();
  
  if (SO_KEY_PRESS_EVENT(event, D)) {
    sv->showVolumes = false;
    keyboardCallback->setHandled();
    ONDEBUG(printf("Shadow volumes: hide\n"));
  }
  else if (SO_KEY_PRESS_EVENT(event, S)) {    
    sv->showVolumes = true;
    keyboardCallback->setHandled();
    ONDEBUG(printf("Shadow volumes: show\n"));
  }
}


//----------------------------------------------------------------------------
// Update shadow volumes when the light position changes
//
void lightMoveCallback (void * userdata, SoSensor *)
{
  ShadowVolume * sv = (ShadowVolume *) userdata;  
  sv->clearShadowVolumeNode();
  sv->createShadowVolumes();
}


//----------------------------------------------------------------------------
// Rendering preset CB
//
static void presetCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("presetCB\n"));

    // depth
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glEnable(GL_CULL_FACE);    
    
    // Clear all buffers
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glPushAttrib(GL_ENABLE_BIT);

    glEnable(GL_LIGHTING);

    SceneLights * lights = (SceneLights *) userdata;
    if (userdata != NULL) lights->lightsOFF();

  }
} 


//----------------------------------------------------------------------------
// Rendering ++ CB
//
static void incrementStencilCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("incrementStencilCB\n"));

    glPopAttrib();

    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glColorMask(0, 0, 0, 0);
    glShadeModel(GL_FLAT);
    glDepthMask(0);
    glDepthFunc(GL_LESS);
    glEnable(GL_STENCIL_TEST);

    if (zfail) {
      // z-fail
      //Increment stencil buffer for back face depth fail
      glStencilFunc(GL_ALWAYS, 0, ~0);
      glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);
      glCullFace(GL_FRONT);
    } else {
      // z-pass
      //Increment stencil buffer for front face depth pass
      glStencilFunc(GL_ALWAYS, 0, ~0);
      glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
      glCullFace(GL_BACK);
    }

  }
} 


//----------------------------------------------------------------------------
// Rendering -- CB
//
static void decrementStencilCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("decrementStencilCB\n"));

    if (zfail) {
      // z-fail
      // Decrement stencil buffer for back face depth pass
      glStencilOp(GL_KEEP, GL_DECR, GL_KEEP);
      glCullFace(GL_BACK);
    } else {
      // z-pass
      //Decrement stencil buffer for back face depth pass
      glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
      glCullFace(GL_FRONT);
    }

  }
} 


//----------------------------------------------------------------------------
// Rendering lit CB
//
static void litStencilCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {

    ONDEBUG(printf("litStencilCB\n"));

    glPopAttrib();

    //
    // Draw lit where stencil == 0
    //

    glPushAttrib(GL_ENABLE_BIT);

    SceneLights * lights = (SceneLights *) userdata;
    if (userdata != NULL) lights->lightsON();

    glEnable(GL_LIGHTING);

    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
    glDepthFunc(GL_EQUAL);
    glStencilFunc(GL_EQUAL, 0, ~0);
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    glEnable(GL_STENCIL_TEST);
    
  }
} 


//----------------------------------------------------------------------------
// Rendering finish CB
//
static void finishStencilCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("wiredStencilCB\n"));
    
    glDisable(GL_LIGHTING);

    glDisable(GL_STENCIL_TEST);

    glPopAttrib();
        
    glPushAttrib(GL_ENABLE_BIT);

    glDepthMask(GL_TRUE);
    glDepthFunc(GL_LEQUAL);

    glEnable(GL_LIGHTING);
    // OpenGL light for wired shadow volumes
    GLfloat lpos[] = {0.0, 0.0, 0.0, 0.0};  	
    GLfloat lcolor[] = {1.0, 0.0, 1.0, 0.0};
        
    glLightfv(GL_LIGHT1, GL_POSITION, lpos);
    glLightfv(GL_LIGHT1, GL_AMBIENT, lcolor);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, lcolor);
    glLightfv(GL_LIGHT1, GL_SPECULAR, lcolor);
    glEnable(GL_LIGHT1);

    glEnable(GL_COLOR_MATERIAL);
    
  }
} 


//----------------------------------------------------------------------------
// Rendering's done CB
//
static void doneCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("doneCB\n"));

    glDisable(GL_LIGHT1);
    glDisable(GL_LIGHTING);    

    glPopAttrib();
  }
} 


//----------------------------------------------------------------------------
// Creates the complete scene with objects
//
SoSeparator* createScene(int argc, char **argv)
{
  int i = 1;

  // Print usage
  if (argc > 1 && (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "--help") == 0)) {
    printf("Usage: %s [-h|--help] [-demo int] [-zfail int]\n"
           "-h, --help        this screen\n"
           "-demo int         possible demo scenes [1, 2, 3]\n"
           "-zfail int        use z-fail algorithm (z-pass default)\n"
           "                  set -zfail 1 to use z-fail\n",
           argv[0]);
    exit(EXIT_SUCCESS);
  }
  
  getParam<int>(argc, argv, i, "-demo", demo);
  getParam<int>(argc, argv, i, "-zfail", zfail);

  // Final check
  if (i != argc) {
    printf("Unknown argument %s.\n", argv[i]);
    exit(EXIT_FAILURE);
  }

  printf("%s %s\n", APPNAME, VERSION);
  printf("Press 's' to show shadow volumes\n");
  printf("Press 'd' to hide shadow volumes\n");
  printf("Use PointLightManipulator to move the light\n\n");

  if (zfail) ONDEBUG(printf("Using z-fail\n"));
  else ONDEBUG(printf("Using z-pass\n"));

  ONDEBUG(printf("Demo scene #%d\n",demo));

  // Ambient light intensity
  SoEnvironment *env = new SoEnvironment;
  env->ambientIntensity.setValue(0.3f);

  // Light 1
  SoPointLightManip *pointLight = new SoPointLightManip;
  pointLight->location.setValue(10.0f, -10.0f, 0.0f); 
  pointLight->intensity.setValue(0.5f);
  pointLight->on = true;
/*
  // Light 2
  SoPointLight *pointLight2 = new SoPointLight;
  pointLight2->location.setValue(-20.0f, 10.0f, -5.0f); 
  pointLight2->color.setValue(0.1f, 0.3f, 0.3f);
  pointLight2->intensity.setValue(0.5f);
  pointLight2->on = true;
*/
  // One root to rule them all :)
  SoSeparator *superRoot = new SoSeparator;
  superRoot->ref();

  // main node of the scene
  SoSeparator *sceneRoot = new SoSeparator;
  sceneRoot->ref();  

  // create light manager
  SceneLights * lights = new SceneLights(sceneRoot); 

  // Root node for all shadow volumes must be under this node
  SoSeparator * shadowRoot = new SoSeparator;

  // all shadow volumes must be under this node
  SoSeparator * wireRoot = new SoSeparator;

  // Camera
  SoPerspectiveCamera * camera = new SoPerspectiveCamera;
  camera->position = SbVec3f(0.f, -80.f, 20.f);
  camera->pointAt(SbVec3f(0, 0, -10));
  // Near & Far planes
  camera->nearDistance = 1.f;
  camera->farDistance = 500.f;

  //
  // Here is the place to create objects of the scene
  //


  //
  // Space Ship model
  //

  // First, create a parent node for a new object
  SoSeparator * shipParent = new SoSeparator;  
  SoSeparator * cubeParent = new SoSeparator;  

  // Move the object
  SoTranslation *trans2 = new SoTranslation;
  trans2->translation.setValue(0.f, 10.f, -25.0f);
  shipParent->addChild(trans2);

  SoRotationXYZ *rotX = new SoRotationXYZ;
  rotX->axis = SoRotationXYZ::X;
  rotX->angle.setValue(90.0f);
  shipParent->addChild(rotX);

  //
  // Load models
  //

  // simple_ship.iv //
  char * filename = NULL;
  filename = "simple_ship.iv";
 
  SoInput in2;
  if (!in2.openFile(filename)) {
    printf("Can not open input file.\n");
    exit(-1);
  }

  SoSeparator *ship = SoDB::readAll(&in2);
  if (ship == NULL) {
    printf("Can not read input file.\n");
    exit(-1);
  }

  shipParent->addChild(ship);

  // box.iv //
  filename = "cube.iv";
 
  SoInput in1;
  if (!in1.openFile(filename)) {
    printf("Can not open input file.\n");
    exit(-1);
  }

  SoSeparator * cube = SoDB::readAll(&in1);
  if (cube == NULL) {
    printf("Can not read input file.\n");
    exit(-1);
  }

  cubeParent->addChild(cube);


  //
  // Scene definition
  // 
  // You can change this code to place your own objects into the scene

  if (demo == 1) {
    // demo scene 1: box with rocket and a cube above
    sceneRoot->addChild(pointLight);
    sceneRoot->addChild(createSurface(SURFACE_DEPTH, SbColor(0.7f, 0.6f, 0.5f)));
    sceneRoot->addChild(shipParent);
    sceneRoot->addChild(cubeParent);      
    //
    lights->addLight(pointLight);
  } else { 
    // demo scene 2: rocket above terrain    
    sceneRoot->addChild(pointLight);
    sceneRoot->addChild(shipParent);
    sceneRoot->addChild(createTerrain(argc, argv));
    //
    lights->addLight(pointLight);
  }


  // Create shadow volumes for scene objects
  // Note: By reason to show both demo scenes in one aplication,
  // shadow volume 'sv1' is creeated for cube in scene 2 eventhough cube 
  // is not in its scene graph. There will be an error printed, please, 
  // ignore it.
  ShadowVolume * sv2 = new ShadowVolume(sceneRoot, shadowRoot, wireRoot, ship);
  ShadowVolume * sv1 = new ShadowVolume(sceneRoot, shadowRoot, wireRoot, cube);
 
  // Keyboard Callbacks
  SoEventCallback * myEventCallback = new SoEventCallback;
  myEventCallback->addEventCallback(SoKeyboardEvent::getClassTypeId(),
                                    myKeyPressCallback, sv2);
  if (demo == 1) {
    myEventCallback->addEventCallback(SoKeyboardEvent::getClassTypeId(),
                                    myKeyPressCallback, sv1);
  }

  sceneRoot->addChild(myEventCallback);

  // Light sensors
  SoNodeSensor * lightSensor1 = new SoNodeSensor(lightMoveCallback, sv2);
  lightSensor1->attach((SoNode *) pointLight);  
  
  if (demo == 1) {
    SoNodeSensor * lightSensor2 = new SoNodeSensor(lightMoveCallback, sv1);
    lightSensor2->attach((SoNode *) pointLight);    
  }

  //
  // Rendering callbacks
  //
  // DO NOT EDIT FOLLOWING CODE

  SoCallback * presetCallback = new SoCallback;
  presetCallback->setCallback(presetCB, lights);
  
  SoCallback * incrementStencilCallback = new SoCallback;
  incrementStencilCallback->setCallback(incrementStencilCB, lights);

  SoCallback * decrementStencilCallback = new SoCallback;
  decrementStencilCallback->setCallback(decrementStencilCB, lights);

  SoCallback * litStencilCallback = new SoCallback;
  litStencilCallback->setCallback(litStencilCB, lights);

  SoCallback * finishStencilCallback = new SoCallback;
  finishStencilCallback->setCallback(finishStencilCB, lights);

  SoCallback * doneCallback = new SoCallback;
  doneCallback->setCallback(doneCB, lights);

  //
  // Put everything together
  // 
  // Create a super root with callbacks and add scene, shadows and 
  // callbacks to switch stencil buffer settings for rendering shadows.
  //

  superRoot->addChild(camera);
  superRoot->addChild(presetCallback);
  superRoot->addChild(sceneRoot);
  superRoot->addChild(incrementStencilCallback);
  superRoot->addChild(shadowRoot);
  superRoot->addChild(decrementStencilCallback);
  superRoot->addChild(shadowRoot);
  superRoot->addChild(litStencilCallback);
  superRoot->addChild(sceneRoot);
  superRoot->addChild(finishStencilCallback);
  superRoot->addChild(wireRoot);
  superRoot->addChild(doneCallback);

  // 

  sceneRoot->unrefNoDelete();
  superRoot->unrefNoDelete();
  return superRoot;
}


