/*****************************************************************************\
 *
 * shadow_volume.cpp
 *
 * Algorithm for shadows ganeration in a scene
 * 
 *  Description: Shadow volume class implementation
 *
 *      Authors: Tom Burian (xburia10 _AT stud.fit.vutbr.cz)
 *               
 *    Thanks to: PCJohn (peciva _AT fit.vutbr.cz)
 * 
 * Contributors:
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/

#include "shadow_volume.h"
#include "shadows_win.h"
#include <windows.h>
#include <GL/gl.h>

//----------------------------------------------------------------------------
// Constructor
//
ShadowVolume::ShadowVolume(SoSeparator * root, SoSeparator * shadow,  SoSeparator * wire, SoSeparator * object)
 : sceneRootNode(root),
   shadowRootNode(shadow),
   wireRootNode(wire),
   modelNode(object),
   svColor(SbColor(0.3f, 0.8f, 0.3f))

{
  farDepth = 100.f;
  stripsCnt = 0;
  created = false;
  showVolumes = false;

  isModelInGraph = findTransformationMatrix();
  if (!isModelInGraph) printError("Cannot find your model in the scene graph.");

  bool hasCoords = findCoordinates();
  bool hasNormals = findNormals();
  bool hasStrips = findStrips();

  shadowVolumeNode = (new SoSeparator);
  shadowRootNode->addChild(shadowVolumeNode);

  wireVolumeNode = (new SoSeparator);
  wireRootNode->addChild(wireVolumeNode);

  //
  this->createShadowVolumes();
}


//----------------------------------------------------------------------------
// Looks for model's transformation matrix
//
bool ShadowVolume::findTransformationMatrix()
{
  SoSearchAction * sa = new SoSearchAction;
  sa->setNode(modelNode);
  sa->setSearchingAll(FALSE);
  sa->apply(sceneRootNode);

  SoPath * nodePath = sa->getPath();
  //assert(nodePath != NULL);
  if ( nodePath == NULL )
    return false;

  const SbViewportRegion vpr;
  SoGetMatrixAction * getmatrixaction = new SoGetMatrixAction(vpr);
  getmatrixaction->apply(nodePath);
  trMatrix = getmatrixaction->getMatrix();

  return true;
}


//----------------------------------------------------------------------------
// Looks for model's normals
//
bool ShadowVolume::findNormals()
{
  SoSearchAction * sa = new SoSearchAction;
  sa->setType(SoNormal::getClassTypeId());
  sa->setSearchingAll(FALSE);
  sa->apply(modelNode);

  SoPath * normalPath = sa->getPath();
  assert(normalPath != NULL);
  if ( normalPath != NULL ) {
    normalNode = (SoNormal *)normalPath->getTail();
    return true;
  } else {
    // not found -> generate own normals?
  }

  return false;
}


//----------------------------------------------------------------------------
// Looks for model's coordinates
//
bool ShadowVolume::findCoordinates()
{
  SoSearchAction * sa = new SoSearchAction;
  sa->setType(SoCoordinate3::getClassTypeId());
  sa->setSearchingAll(FALSE);
  sa->apply(modelNode);

  SoPath * coordsPath = sa->getPath();
  assert(coordsPath != NULL);
  if ( coordsPath != NULL ) {
    coordNode = (SoCoordinate3 *)coordsPath->getTail();
    return true;
  } else {
    // coords has not been fount
  }

  return false;
}


//----------------------------------------------------------------------------
// Looks for model's strip sets
//
bool ShadowVolume::findStrips()
{
  stripsCnt = 0;
  int i = 0;  // actual strip set

  // Finds a stripset node
  SoSearchAction * sa = new SoSearchAction;
  sa->setType(SoShape::getClassTypeId());
  sa->setInterest(SoSearchAction::ALL);
  sa->setSearchingAll(TRUE);
  sa->apply(modelNode);

  const SoPathList & pl = sa->getPaths();

  int pathLength = pl.getLength();

  // Find TriangleStripSets 
  for ( i=0; i < pathLength ; i++) {
    SoFullPath * fp = (SoFullPath *)pl[i];
    assert (fp != NULL );

    if (fp->getTail()->getTypeId() == SoIndexedTriangleStripSet::getClassTypeId()) {
      SoIndexedTriangleStripSet * stripNode = (SoIndexedTriangleStripSet *)fp->getTail();
      stripsCnt++;

    } // if
  } // for

  stripNodes = new SoIndexedTriangleStripSet *[stripsCnt];

  for ( i=0; i < pathLength ; i++) {
    SoFullPath * fp = (SoFullPath *)pl[i];
    assert (fp != NULL );
    if (fp->getTail()->getTypeId() == SoIndexedTriangleStripSet::getClassTypeId()) {
      SoIndexedTriangleStripSet * stripNode = (SoIndexedTriangleStripSet *)fp->getTail();

    stripNodes[i] = stripNode;

    } // if
  } // for

  return stripsCnt > 0;
}


//----------------------------------------------------------------------------
// Destructor
//
ShadowVolume::~ShadowVolume()
{
  delete [] stripNodes;
}


//----------------------------------------------------------------------------
// Removes all children of shadowVolumeNode
// Shadow volumes must be recalculated when light position
// or object position has changed
bool ShadowVolume::clearShadowVolumeNode()
{
  // Delete shadow volumes
  if ( shadowVolumeNode != NULL )
    shadowVolumeNode->removeAllChildren();

  // Delete wire models
  if ( wireVolumeNode != NULL )
    wireVolumeNode->removeAllChildren();

  return true;
}


//----------------------------------------------------------------------------
// Writes generated shadow node scene graph into a file
//
void ShadowVolume::writeShadowNode(const char *filename)
{
  SoOutput output;
  if ( created ) {
    output.openFile(filename);
    SoWriteAction * writer = new SoWriteAction(&output);
    writer->apply(shadowVolumeNode);
    delete writer;
    fprintf(stdout,"Shadows' scene graph was written in '%s'.\n\n", filename);
  } else {
    fprintf(stderr,"Error: Could not write shadows' scene graph.\nShadow volumes have not been generated yet.\n\n");
  }
}


//----------------------------------------------------------------------------
// Creates shadow volumes for all lights in the scene
//
void ShadowVolume::createShadowVolumes()
{
  // if model was not fount in the scene graph, unable to 
  // create shadows
  if (isModelInGraph) {

    SoMaterial *material = new SoMaterial;
    material->ambientColor.setValue(svColor);
    material->diffuseColor.setValue(svColor);
    shadowVolumeNode->addChild(material);

    // Finds all lights
    SoSearchAction sa;
    sa.setType(SoLight::getClassTypeId());
    sa.setSearchingAll(TRUE);
    sa.setInterest(SoSearchAction::ALL);

    sa.apply(sceneRootNode);
    const SoPathList & pl = sa.getPaths();

    // Create shadow volumes of the object for all lights
    for (int i=0; i < pl.getLength(); i++) {
      SoFullPath * fp = (SoFullPath *)pl[i];
      if (fp->getTail()->getTypeId() == SoPointLightManip::getClassTypeId() || 
          fp->getTail()->getTypeId() == SoPointLight::getClassTypeId()) { 
        SoPointLight * plight = (SoPointLight *)fp->getTail();
        createShadowVolume(plight);
      }
    } // for

    created = true;
  } // is in the scene graph
}


//----------------------------------------------------------------------------
// Creates shadow volume for one light
//
void ShadowVolume::createShadowVolume(SoPointLight * light)
{
  int i = 0;  // actual strip set

  // For all strip sets
  for ( i=0; i < stripsCnt; i++ ) {

    // Get one and generate SV for its triangles
    SoIndexedTriangleStripSet * stripNode = stripNodes[i];

    int itn = 0;  // index to normal indices array
    int j = 2; // actual triangle in the strip set, vertices: j-2, j-1 and j

    float vysl = 0.0f;

    SbVec3f normal;

    //
    // Create a shadow volume for each light-facing triangle
    // Goes over all triangles in the triangleStripSet
    while ( j < stripNode->coordIndex.getNum() ) {

      while ( stripNode->coordIndex[j] != -1 ) {

        // Debug: Check if we are in the range
        int b = stripNode->normalIndex.getNum();
        if ( b < itn ) {
          printError("Normal index out of range.\n\n");
          exit(1);
        }

        // Get the normal vector
        // Todo: compute my own 
        int nindex = stripNode->normalIndex[itn];
        normal = normalNode->vector[nindex];

        // Transformation of a vertex and normal vector to global coordinates
        SbVec3f x(coordNode->point[stripNode->coordIndex[j-2]]);
        SbVec3f normEnd(x+1*normal);  // endpoint of the normal vector
        // Vertex
        trMatrix.multVecMatrix(x, x);
        // New normal vector
        trMatrix.multVecMatrix(normEnd, normEnd);
        normal = normEnd - x;
        normal.normalize();

        // Vector from the vertex to light source
        SbVec3f lv(light->location.getValue()-x);
        // Dot product of normal and light vector
        vysl = normal.dot(lv);

        if ( vysl <= 0.0f ) {
          // Not facing light

        } else {
          // Is light-facing, part of the silhouette

          // Create envelope of the shadow volume

          // Coords
          SoCoordinate3 *envCoords = new SoCoordinate3;
          envCoords->point.setNum(2*3+2); // 8 = 2 * #edges + 2 (closed)

          SbVec3f *c = envCoords->point.startEditing();

          for ( int k=0; k <= 3 ;k++ ) {
            int t = j-k%3;
            // To global coordinates
            SbVec3f xx(coordNode->point[stripNode->coordIndex[j-k%3]]);
            trMatrix.multVecMatrix(xx, xx);
            SbVec3f xxFar(farPoint(light->location.getValue(), xx));

            *(c++) = xx;
            *(c++) = xxFar;  // Far point
            
          }
          envCoords->point.finishEditing();

          // Add nodes to scene graph
          shadowVolumeNode->addChild(envCoords);

          SoIndexedTriangleStripSet * envStrip = new SoIndexedTriangleStripSet;
          int32_t indeces[24] = {0,1,2,-1,2,1,3,-1, 2,3,4,-1,4,3,5,-1, 4,5,6,-1,6,5,7,-1};
          envStrip->coordIndex.setValues(0,24,indeces);

          shadowVolumeNode->addChild(envStrip);

          // Near & far cap of the volume
          SoIndexedTriangleStripSet * ncapStrip = new SoIndexedTriangleStripSet;
          int32_t ncapIndeces[4] = {0,2,4,-1};
          ncapStrip->coordIndex.setValues(0,4,ncapIndeces);
          shadowVolumeNode->addChild(ncapStrip);

          SoIndexedTriangleStripSet * fcapStrip = new SoIndexedTriangleStripSet;
          int32_t fcapIndeces[4] = {1,3,5,-1};
          fcapStrip->coordIndex.setValues(0,4,fcapIndeces);
          shadowVolumeNode->addChild(fcapStrip);


          // Add wire volumes

          if (this->showVolumes) {

            // Add wire node to shadow scene graph
            wireVolumeNode->addChild(envCoords);

            for (int k=0; k<3; k++) {
              SoIndexedLineSet * line = new SoIndexedLineSet;
              int32_t lineIndeces[3];
              lineIndeces[0] = 2*k;
              lineIndeces[1] = 2*k+1;
              lineIndeces[2] = -1;
              line->coordIndex.setValues(0,3,lineIndeces);
              wireVolumeNode->addChild(line);
            }              

            SoIndexedLineSet * flinecap = new SoIndexedLineSet;
            int32_t flinecapInd[5] = {1,3,5,1,-1};
            flinecap->coordIndex.setValues(0,5,flinecapInd);
            wireVolumeNode->addChild(flinecap);

          } // end wires

        }
        
        // End of a strip, move iterators
        itn++;
        j++;

      } // while

      j += 3;  // end of strip (-1) => move to next one

    } // while

  } // one stripset end
}



//----------------------------------------------------------------------------
// Prints information about shadow volume generation
//
void ShadowVolume::printInfo(const char *msg)
{
  fprintf(stdout,"SV:\t%s\n\n",msg);
}


//----------------------------------------------------------------------------
// Error printing method
//
void ShadowVolume::printError(const char *msg)
{
  fprintf(stdout,"SV Error:\t%s\n\n",msg);
}


//----------------------------------------------------------------------------
// Computes shadow volume far plane point (far cap)
//
SbVec3f ShadowVolume::farPoint(SbVec3f lightPos, SbVec3f point)
{
  SbVec3f farPoint(point);
  SbVec3f l2pVec(point-lightPos);
  l2pVec.normalize();
  // farPoint = lightPos + farDepth * l2pVec;
  farPoint = point + farDepth * l2pVec;

  return farPoint;
}


